""" Functions to convert objects """

import logging
from datetime import datetime, timezone
from typing import List

import pandas as pd
from sqlalchemy.orm import Session

from nowcasting_datamodel.models import (
    Forecast,
    ForecastSQL,
    ForecastValue,
    ForecastValueSevenDaysSQL,
)
from nowcasting_datamodel.read.read import (
    get_latest_input_data_last_updated,
    get_location,
    get_model,
)

logger = logging.getLogger()


def convert_list_forecast_value_seven_days_sql_to_list_forecast(
    forecast_values_sql=List[ForecastValueSevenDaysSQL],
) -> List[Forecast]:
    """
    Convert a list of sql ForecastValueSevenDaysSQL to pydantic Forecast objects

    :param forecast_values_sql: forecast value objects
    :return: list of Forecast pydantic objects
    """

    # dictionary of forecasts by gsp_id
    forecasts_by_gsp = {}
    for forecast_value_sql in forecast_values_sql:
        gsp_id = forecast_value_sql.forecast.location.gsp_id

        forecast_value: ForecastValue = ForecastValue.from_orm(forecast_value_sql)

        if gsp_id in forecasts_by_gsp.keys():
            forecasts_by_gsp[gsp_id].forecast_values.append(forecast_value)
        else:
            forecast = ForecastSQL(
                model=forecast_value_sql.forecast.model,
                forecast_creation_time=forecast_value_sql.forecast.forecast_creation_time,
                location=forecast_value_sql.forecast.location,
                input_data_last_updated=forecast_value_sql.forecast.input_data_last_updated,
                forecast_values=[forecast_value_sql],
                historic=forecast_value_sql.forecast.historic,
            )
            forecast = Forecast.from_orm(forecast)
            forecasts_by_gsp[gsp_id] = forecast

    forecasts = [forecast for forecast in forecasts_by_gsp.values()]

    return forecasts


def convert_df_to_national_forecast(
    forecast_values_df: pd.DataFrame, session: Session, model_name: str, version: str
) -> ForecastSQL:
    """
    Make a ForecastSQL object from a dataframe.

    :param forecast_values_df: Dataframe containing
        -- target_datetime_utc
        -- forecast_mw
        -- (Optional) forecast_mw_plevel_10
        -- (Optional) forecast_mw_plevel_90
    :param: session: database session
    :param: model_name: the name of the model
    :param: version: the version of the model
    :return: forecast object
    """

    logger.debug("Converting dataframe to National Forecast")

    assert "target_datetime_utc" in forecast_values_df.columns
    assert "forecast_mw" in forecast_values_df.columns

    # get last input data
    input_data_last_updated = get_latest_input_data_last_updated(session=session)

    # get model name
    model = get_model(name=model_name, version=version, session=session)

    # get location
    location = get_location(session=session, gsp_id=0)

    # make forecast values
    forecast_values = []
    for i, forecast_value in forecast_values_df.iterrows():
        # add timezone
        target_time = forecast_value.target_datetime_utc.replace(tzinfo=timezone.utc)
        forecast_value_sql = ForecastValue(
            target_time=target_time,
            expected_power_generation_megawatts=forecast_value.forecast_mw,
        ).to_orm()
        forecast_value_sql.adjust_mw = 0.0

        forecast_value_sql.properties = {}
        if "forecast_mw_plevel_10" in forecast_values_df.columns:
            forecast_value_sql.properties["10"] = forecast_value.forecast_mw_plevel_10

        if "forecast_mw_plevel_90" in forecast_values_df.columns:
            forecast_value_sql.properties["90"] = forecast_value.forecast_mw_plevel_90

        forecast_values.append(forecast_value_sql)

    # make forecast object
    forecast = ForecastSQL(
        model=model,
        forecast_creation_time=datetime.now(tz=timezone.utc),
        location=location,
        input_data_last_updated=input_data_last_updated,
        forecast_values=forecast_values,
        historic=False,
    )

    return forecast



""" Pydantic and Sqlalchemy models for the database

The following class are made
1. ForecastValue objects, specific values of a forecast and time
2. Forecasts, a forecast that is made for one gsp, for several time steps into the future

"""

import logging
from datetime import datetime
from typing import List

import numpy as np
from pydantic import Field, validator
from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Float,
    ForeignKey,
    Index,
    Integer,
    event,
    func,
)
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import DeclarativeMeta, declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship
from sqlalchemy.sql.ddl import DDL

from nowcasting_datamodel.models.base import Base_Forecast
from nowcasting_datamodel.models.gsp import Location
from nowcasting_datamodel.models.models import InputDataLastUpdated, MLModel
from nowcasting_datamodel.models.utils import CreatedMixin, EnhancedBaseModel
from nowcasting_datamodel.utils import datetime_must_have_timezone

logger = logging.getLogger(__name__)

"""
Tried to follow the example here
https://stackoverflow.com/questions/61545680/postgresql-partition-and-sqlalchemy
"""


class PartitionByMeta(DeclarativeMeta):
    """Partition table meta object"""

    def __new__(cls, clsname, bases, attrs, *, partition_by, partition_type: str = "RANGE"):
        """Make new partition"""

        @classmethod
        def get_partition_name(cls_, suffix):
            """Get the name of the partition table"""
            return f"{cls_.__tablename__}_{suffix}"

        @classmethod
        def create_partition(
            cls_, suffix, year_month_end, subpartition_by=None, subpartition_type=None
        ):
            """Create new partitions"""
            if suffix not in cls_.partitions:
                partition = PartitionByMeta(
                    f"{clsname}{suffix}",
                    bases,
                    {"__tablename__": cls_.get_partition_name(suffix)},
                    partition_type=subpartition_type,
                    partition_by=subpartition_by,
                )

                partition.__table__.add_is_dependent_on(cls_.__table__)

                event.listen(
                    partition.__table__,
                    "after_create",
                    DDL(
                        # For non-year ranges, modify the FROM and TO below
                        # LIST: IN ('first', 'second');
                        # RANGE: FROM ('{key}-01-01') TO ('{key+1}-01-01')
                        f"""
                        ALTER TABLE {cls_.__tablename__}
                        ATTACH PARTITION {partition.__tablename__}
                        FOR VALUES FROM ('{suffix}-01') TO ('{year_month_end}-01');
                        """
                    ),
                )

                cls_.partitions[suffix] = partition

            return cls_.partitions[suffix]

        if partition_by is not None:
            attrs.update(
                {
                    "__table_args__": attrs.get("__table_args__", ())
                    + (dict(postgresql_partition_by=f"{partition_type.upper()}({partition_by})"),),
                    "partitions": {},
                    "partitioned_by": partition_by,
                    "get_partition_name": get_partition_name,
                    "create_partition": create_partition,
                }
            )

        return super().__new__(cls, clsname, bases, attrs)


class ForecastValueSQLMixin(CreatedMixin):
    """One Forecast of generation at one timestamp

    This Mixin is used to create partition tables
    """

    uuid = Column(UUID, primary_key=True, server_default=func.gen_random_uuid())
    target_time = Column(DateTime(timezone=True), nullable=False, primary_key=True)
    expected_power_generation_megawatts = Column(Float(precision=6))
    adjust_mw = Column(Float, default=0.0)
    # this can be used to store any additional information about the forecast, like p_levels.
    # Want to keep it as json so that we can store different properties for different forecasts
    properties = Column(MutableDict.as_mutable(JSON), nullable=True)

    @declared_attr
    def forecast_id(self):
        """Link with Forecast table"""
        return Column(Integer, ForeignKey("forecast.id"))


class ForecastValueSQL(
    ForecastValueSQLMixin, Base_Forecast, metaclass=PartitionByMeta, partition_by="target_time"
):
    """One Forecast of generation at one timestamp

    This Mixin is used to creat partition tables
    """

    __tablename__ = "forecast_value"

    __table_args__ = (
        Index(
            "ix_forecast_value_forecast_id",  # Index name
            "forecast_id",  # Columns which are part of the index
        ),
        Index(
            "ix_forecast_value_created_utc",  # Index name
            "created_utc",  # Columns which are part of the index
        ),
        Index(
            "ix_forecast_value_target_time",  # Index name
            "target_time",  # Columns which are part of the index
        ),
    )

    forecast = relationship("ForecastSQL", back_populates="forecast_values")


def get_partitions(start_year: int, start_month: int, end_year: int, end_month: int):
    """Make partitions"""
    partitions = []
    for year in range(start_year, end_year + 1):
        if year != start_year:
            start_month = 1
        if year == end_year:
            end_month_loop = end_month
        else:
            end_month_loop = 13

        for month in range(start_month, end_month_loop):
            if month == 12:
                year_end = year + 1
                month_end = 1
            else:
                year_end = year
                month_end = month + 1

            if month < 10:
                month = f"0{month}"
            if month_end < 10:
                month_end = f"0{month_end}"

            partitions.append(
                ForecastValueSQL.create_partition(f"{year}_{month}", f"{year_end}_{month_end}")
            )

    return partitions


def make_partitions(start_year: int, start_month: int, end_year: int):
    """
    Make partitions

    :param start_year: year to start
    :param start_month: month to end
    :param end_year: end year (exclusive)
    """
    for year in range(start_year, end_year):
        if year != start_year:
            start_month = 1

        for month in range(start_month, 13):
            if month == 12:
                year_end = year + 1
                month_end = 1
            else:
                year_end = year
                month_end = month + 1

            if month < 10:
                month = f"0{month}"
            if month_end < 10:
                month_end = f"0{month_end}"

            class ForecastValueYearMonth(ForecastValueSQLMixin, Base_Forecast):
                __tablename__ = f"forecast_value_{year}_{month}"

                __table_args__ = (
                    Index(
                        f"forecast_value_{year}_{month}_created_utc_idx",  # Index name
                        "created_utc",  # Columns which are part of the index
                    ),
                    Index(
                        f"forecast_value_{year}_{month}_target_time_idx",  # Index name
                        "target_time",  # Columns which are part of the index
                    ),
                    Index(
                        f"forecast_value_{year}_{month}_forecast_id_idx",  # Index name
                        "forecast_id",  # Columns which are part of the index
                    ),
                )

            ForecastValueYearMonth.__table__.add_is_dependent_on(ForecastValueSQL.__table__)

            event.listen(
                ForecastValueYearMonth.__table__,
                "after_create",
                DDL(
                    f"ALTER TABLE forecast_value ATTACH PARTITION forecast_value_{year}_{month} "
                    f"for VALUES FROM ('{year}-{month}-01') TO ('{year_end}-{month_end}-01');"
                ),
            )


make_partitions(2022, 8, 2025)


# legacy table, this means migration still work
class ForecastValueOld(ForecastValueSQLMixin, Base_Forecast):
    """Old ForecastValue table"""

    __tablename__ = "forecast_value_old"

    __table_args__ = (
        Index(
            "ix_forecast_value_forecast_id_old",  # Index name
            "forecast_id",  # Columns which are part of the index
        ),
        Index(
            "ix_forecast_value_target_time_old",  # Index name
            "target_time",  # Columns which are part of the index
        ),
    )


class ForecastValueLatestSQL(Base_Forecast, CreatedMixin):
    """One Forecast of generation at one timestamp

    Thanks for the help from
    https://www.johbo.com/2016/creating-a-partial-unique-index-with-sqlalchemy-in-postgresql.html
    """

    __tablename__ = "forecast_value_latest"

    # add a unique condition on 'gsp_id',  'target_time' and 'model_id'
    __table_args__ = (
        Index(
            "uix_1",  # Index name
            "gsp_id",  # Columns which are part of the index
            "target_time",
            "model_id",
            unique=True,
            postgresql_where=Column("is_primary"),  # The condition
        ),
    )

    target_time = Column(DateTime(timezone=True), index=True, primary_key=True)
    expected_power_generation_megawatts = Column(Float(precision=6))
    gsp_id = Column(Integer, index=True, primary_key=True)
    model_id = Column(Integer, index=True, primary_key=True, default=-1)
    is_primary = Column(Boolean, default=True)
    adjust_mw = Column(Float, default=0.0)
    # this can be used to store any additional information about the forecast, like p_levels.
    # Want to keep it as json so that we can store different properties for different forecasts
    properties = Column(MutableDict.as_mutable(JSON), nullable=True)

    forecast_id = Column(Integer, ForeignKey("forecast.id"), index=True)
    forecast_latest = relationship("ForecastSQL", back_populates="forecast_values_latest")


class ForecastValue(EnhancedBaseModel):
    """One Forecast of generation at one timestamp"""

    target_time: datetime = Field(
        ...,
        description="The target time that the forecast is produced for",
    )
    expected_power_generation_megawatts: float = Field(
        ..., ge=0, description="The forecasted value in MW"
    )

    expected_power_generation_normalized: float = Field(
        None, ge=0, description="The forecasted value divided by the gsp capacity [%]"
    )

    _adjust_mw: float = Field(
        0.0,
        description="The amount that the forecast should be adjusted by, "
        "due to persistence errors. This way we keep the original ML prediction. "
        "The _ at the start means it is not expose in the API",
    )

    # This its better to keep this out of the current pydantic models used by the API.
    # A new pydantic mode can be made that includes the forecast plevels, perhaps in the API.
    _properties: dict = Field(
        None,
        description="Dictionary to hold properties of the forecast, like p_levels. "
        "The _ at the start means it is not expose in the API",
    )

    _normalize_target_time = validator("target_time", allow_reuse=True)(datetime_must_have_timezone)

    def to_orm(self) -> ForecastValueSQL:
        """Change model to ForecastValueSQL"""
        return ForecastValueSQL(
            target_time=self.target_time,
            expected_power_generation_megawatts=self.expected_power_generation_megawatts,
            adjust_mw=self._adjust_mw,
        )

    @classmethod
    def from_orm(cls, obj: ForecastValueSQL):
        """Make sure _adjust_mw is transfered also"""
        m = super().from_orm(obj=obj)

        # this is because from orm doesnt copy over '_' variables.
        # But we don't want to expose this in the API
        default_value = 0.0
        if hasattr(obj, "adjust_mw"):
            adjust_mw = obj.adjust_mw
            if not adjust_mw or np.isnan(adjust_mw):
                adjust_mw = default_value
            m._adjust_mw = adjust_mw
        else:
            m._adjust_mw = default_value

        return m

    def normalize(self, installed_capacity_mw):
        """Normalize forecasts by installed capacity mw"""
        if installed_capacity_mw in [0, None]:
            self.expected_power_generation_normalized = 0
        else:
            self.expected_power_generation_normalized = (
                self.expected_power_generation_megawatts / installed_capacity_mw
            )

        return self

    def adjust(self, limit: float = 0.0):
        """Adjust forecasts"""
        if isinstance(self._adjust_mw, float):
            adjust_mw = self._adjust_mw
            if adjust_mw > limit:
                adjust_mw = limit
            elif adjust_mw < -limit:
                adjust_mw = -limit

            self.expected_power_generation_megawatts = (
                self.expected_power_generation_megawatts - adjust_mw
            )

            if self.expected_power_generation_megawatts < 0:
                self.expected_power_generation_megawatts = 0.0

            if isinstance(self._properties, dict):
                for p_level in ["10", "90"]:
                    if p_level in self._properties.keys():
                        if self._properties[p_level] is not None:
                            self._properties[p_level] -= adjust_mw
                            if self._properties[p_level] < 0:
                                self._properties[p_level] = 0.0

        return self


class ForecastSQL(Base_Forecast, CreatedMixin):
    """Forecast SQL model"""

    __tablename__ = "forecast"

    __table_args__ = (
        Index(
            "idx_forecast_created_utc",  # Index name
            "created_utc",  # Columns which are part of the index
        ),
    )

    id = Column(Integer, primary_key=True)
    forecast_creation_time = Column(DateTime(timezone=True))

    # Two distinuise between
    # 1. a Forecast with some forecast values, for that moment,
    # 2. a Forecast with all the historic latest value.
    # we use this boolean.
    # This make it easier to load the forecast showing historic values very easily
    historic = Column(Boolean, default=False)

    model = relationship("MLModelSQL", back_populates="forecast")
    model_id = Column(Integer, ForeignKey("model.id"), index=True)

    # many (forecasts) to one (location)
    location = relationship("LocationSQL", back_populates="forecast")
    location_id = Column(Integer, ForeignKey("location.id"), index=True)

    # one (forecasts) to many (forecast_value)
    forecast_values = relationship("ForecastValueSQL")
    forecast_values_latest = relationship(
        "ForecastValueLatestSQL", back_populates="forecast_latest"
    )
    forecast_values_last_seven_days = relationship(
        "ForecastValueSevenDaysSQL", back_populates="forecast"
    )

    # many (forecasts) to one (input_data_last_updated)
    input_data_last_updated = relationship("InputDataLastUpdatedSQL", back_populates="forecast")
    input_data_last_updated_id = Column(
        Integer, ForeignKey("input_data_last_updated.id"), index=True
    )

    Index("index_forecast_historic", historic)


class Forecast(EnhancedBaseModel):
    """A single Forecast"""

    location: Location = Field(..., description="The location object for this forecaster")
    model: MLModel = Field(..., description="The name of the model that made this forecast")
    forecast_creation_time: datetime = Field(
        ..., description="The time when the forecaster was made"
    )
    historic: bool = Field(
        False,
        description="if False, the forecast is just the latest forecast. "
        "If True, historic values are also given",
    )
    forecast_values: List[ForecastValue] = Field(
        ...,
        description="List of forecasted value objects. Each value has the datestamp and a value",
    )
    input_data_last_updated: InputDataLastUpdated = Field(
        ...,
        description="Information about the input data that was used to create the forecast",
    )

    _normalize_forecast_creation_time = validator("forecast_creation_time", allow_reuse=True)(
        datetime_must_have_timezone
    )

    def to_orm(self) -> ForecastSQL:
        """Change model to ForecastSQL"""
        return ForecastSQL(
            model=self.model.to_orm(),
            forecast_creation_time=self.forecast_creation_time,
            location=self.location.to_orm(),
            input_data_last_updated=self.input_data_last_updated.to_orm(),
            forecast_values=[forecast_value.to_orm() for forecast_value in self.forecast_values],
            historic=self.historic,
        )

    @classmethod
    def from_orm_latest(cls, forecast_sql: ForecastSQL):
        """Method to make Forecast object from ForecastSQL,

        but move 'forecast_values_latest' to 'forecast_values'
        This is useful as we want the API to still present a Forecast object.
        """
        # do normal transform
        forecast = cls.from_orm(forecast_sql)

        # move 'forecast_values_latest' to 'forecast_values'
        forecast.forecast_values = [
            ForecastValue.from_orm(forecast_value)
            for forecast_value in forecast_sql.forecast_values_latest
        ]

        return forecast

    def normalize(self, adjust: bool = False):
        """Normalize forecasts by installed capacity mw"""

        self.forecast_values = [
            forecast_value.normalize(self.location.installed_capacity_mw)
            for forecast_value in self.forecast_values
        ]

        return self

    def adjust(self, limit: float = 0.0):
        """Adjust forecasts by adjust values, useful for time persistence errors"""

        print("start main")
        self.forecast_values = [
            forecast_value.adjust(limit=limit) for forecast_value in self.forecast_values
        ]

        return self


class ManyForecasts(EnhancedBaseModel):
    """Many Forecasts"""

    forecasts: List[Forecast] = Field(
        ...,
        description="List of forecasts for different GSPs",
    )

    def normalize(self):
        """Normalize forecasts by installed capacity mw"""
        self.forecasts = [forecast.normalize() for forecast in self.forecasts]

    def adjust(self, limit: float = 0.0):
        """Adjust forecasts"""
        self.forecasts = [forecast.adjust(limit=limit) for forecast in self.forecasts]


class ForecastValueSevenDaysSQL(ForecastValueSQLMixin, Base_Forecast):
    """
    One Forecast of generation at one timestamp

    This table will only save the last week of data.
    """

    __tablename__ = "forecast_value_last_seven_days"

    __table_args__ = (
        Index(
            "idx_forecast_value_last_seven_days_created_utc",  # Index name
            "created_utc",  # Columns which are part of the index
        ),
        Index(
            "idx_forecast_value_last_seven_days_target_time",  # Index name
            "target_time",  # Columns which are part of the index
        ),
    )

    forecast = relationship("ForecastSQL", back_populates="forecast_values_last_seven_days")


Index("ix_forecast_created_utc", ForecastSQL.created_utc.desc())
Index("ix_forecast_value_seven_days_created_utc", ForecastValueSevenDaysSQL.created_utc.desc())
Index("ix_forecast_value_latest_created_utc", ForecastValueLatestSQL.created_utc.desc())



""" Methods for adding adjust values to the forecast"""
import logging
from datetime import datetime, timedelta
from typing import List, Union

import numpy as np
import pandas as pd

from nowcasting_datamodel.models import Forecast, ForecastSQL, MetricValue, MetricValueSQL
from nowcasting_datamodel.read.read_metric import read_latest_me_national

logger = logging.getLogger()


def add_adjust_to_forecasts(forecasts_sql: List[ForecastSQL], session):
    """
    Adjust National Forecast by ME over the last week

    Note that no objects are returned as sqlalchemy objects can be adjust on the spot

    :param forecasts_sql: list of forecasts
    :param session: database sessions
    """
    logger.debug("Adding Adjusts to National forecast")

    # get the national forecast only
    forecast_national = [f for f in forecasts_sql if f.location.gsp_id == 0]

    if len(forecast_national) != 1:
        logger.debug("Could not find single national forecast, tehre fore not adding adjust")

    add_adjust_to_national_forecast(forecast=forecast_national[0], session=session)


def add_adjust_to_national_forecast(forecast: ForecastSQL, session):
    """
    Add adjust to national forecast.

    1. Get latest me results for time_of_day and forecast horizont
    2. Reduce latest me results relevant for now
    3. Add Me to adjust_mw property in ForecastValueSQL

    :param forecast: national forecast
    :param session:
    :return:
    """

    # get the target time and model name
    datetime_now = forecast.forecast_values[0].target_time
    last_datetime = forecast.forecast_values[-1].target_time
    model_name = forecast.model.name

    logger.debug(
        f"Adding adjuster to national forecast for "
        f"{datetime_now} and {model_name}, {last_datetime=}"
    )

    # 1. read metric values
    latest_me = read_latest_me_national(session=session, model_name=model_name)
    if len(latest_me) == 0:
        logger.warning(f"Found no ME values found for {model_name=}")
    else:
        logger.debug(f"Found {len(latest_me)} latest ME values")

    # 2. filter value down to now onwards
    # get the number of hours to go ahead, we've added 1 to make sure we use the last one as well
    hours_ahead = get_forecast_horizon_from_forecast(forecast)
    # change to dataframe
    latest_me_df = reduce_metric_values_to_correct_forecast_horizon(
        latest_me=latest_me, datetime_now=datetime_now, hours_ahead=hours_ahead
    )
    assert len(latest_me_df) > 0

    # 3. add values to forecast_values
    for forecast_value in forecast.forecast_values:
        target_time = forecast_value.target_time
        try:
            # get value from df
            value = latest_me_df[latest_me_df["datetime"] == forecast_value.target_time]
            value = value.iloc[0].value

            # add value to ForecastValueSQL
            forecast_value.adjust_mw = value
            if np.isnan(value):
                logger.debug(
                    f"Found ME value for {target_time} in {latest_me_df}, "
                    f"but it was NaN, therefore adding adjust_mw as 0"
                )
                forecast_value.adjust_mw = 0.0

        except Exception:
            logger.debug(
                f"Could not find ME value for {target_time} in {latest_me_df}, "
                f"therefore adding adjust_mw as 0"
            )
            forecast_value.adjust_mw = 0.0


def get_forecast_horizon_from_forecast(forecast: Union[ForecastSQL, Forecast]):
    """
    Get the number of hours ahead from the forecast

    :param forecast:
    :return:
    """
    datetime_now = forecast.forecast_values[0].target_time
    last_datetime = forecast.forecast_values[-1].target_time
    hours_ahead = int((last_datetime - datetime_now).seconds / 3600) + 1
    hours_ahead += (last_datetime - datetime_now).days * 24

    logger.debug(f"Hours ahead is {hours_ahead}, {last_datetime=}, {datetime_now=} ")

    return hours_ahead


def reduce_metric_values_to_correct_forecast_horizon(
    latest_me: List[MetricValueSQL], datetime_now: datetime, hours_ahead=4
) -> pd.DataFrame:
    """
    Change all latest ME results 2D to 1D array

    Start:
    # time forecast_horizon value
    # 00.00 0  0.1
    # 00.00 30 0.2
    # 00.00 60 0.6
    # .....
    # 00.30 0  0.11
    # 00.30 30 0.21
    # 00.30 60 0.61
    # .....
    # 23.30 0  0.11
    # 23.30 30 0.21
    # 23.30 60 0.61

    Using datetime_now of 04:30

    End:
    # time forecast_horizon value
    # 04.30 0  0.1
    # 05.00 30 0.2
    # 05.30 60 0.6
    # .....

    :param latest_me: list of latest me results for all time_of_day and all forecast_horizons
    :param datetime_now: the dateteim now, so we know where to use forecast_horizons of 0
    :param hours_ahead: how many hours ahead
    :return: dataframe containing
        'datetime'
        'time_of_day'
        'forecast_horizon'
        'value'
    """

    logger.debug(
        f"Reducing metric values to correct forecast horizon {datetime_now=} {hours_ahead=}"
    )

    latest_me_df = pd.DataFrame([MetricValue.from_orm(m).dict() for m in latest_me])
    if len(latest_me_df) == 0:
        # no latest ME values, so just making an empty dataframe
        latest_me_df = pd.DataFrame(columns=["forecast_horizon_minutes", "time_of_day", "value"])

    # Let now big a dataframe of datetimes from now onwards. Lets say the time is 04.30, then
    # time forecast_horizon value
    # 04.30 0  0.1
    # 05.00 30 0.2
    # 05.30 60 0.6
    # .....

    results_df = pd.DataFrame(
        index=pd.date_range(
            start=datetime_now, end=datetime_now + timedelta(hours=hours_ahead), freq="30T"
        ),
    )
    results_df["datetime"] = results_df.index
    results_df["time_of_day"] = results_df.index.time
    results_df["forecast_horizon_minutes"] = range(0, 30 * len(results_df), 30)
    latest_me_df["forecast_horizon_minutes"] = latest_me_df["forecast_horizon_minutes"].astype(int)
    results_df["forecast_horizon_minutes"] = results_df["forecast_horizon_minutes"].astype(int)
    latest_me_df["value"] = latest_me_df["value"].astype(float)

    all_df = pd.merge(
        results_df,
        latest_me_df,
        how="left",
        on=["time_of_day", "forecast_horizon_minutes"],
    )

    all_df = all_df[["datetime", "time_of_day", "value", "forecast_horizon_minutes"]]

    return all_df



""" Method to update latest forecast values """
import logging
from datetime import datetime, timedelta, timezone
from typing import List, Optional

from sqlalchemy import delete, inspect
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm.session import Session

from nowcasting_datamodel import N_GSP
from nowcasting_datamodel.models import ForecastValueSevenDaysSQL
from nowcasting_datamodel.models.forecast import (
    ForecastSQL,
    ForecastValueLatestSQL,
    ForecastValueSQL,
)
from nowcasting_datamodel.read.read import (
    get_latest_forecast,
    get_latest_forecast_for_gsps,
)

logger = logging.getLogger(__name__)


def get_historic_forecast(
    session: Session, forecast: ForecastSQL, model_name: Optional[str] = None
) -> ForecastSQL:
    """
    Get historic forecast

    :param session:
    :param gsp_id:
    :param model_name: the model name to filter on
    :return:
    """

    gsp_id = forecast.location.gsp_id

    forecast_historic = get_latest_forecast(
        session=session, gsp_id=gsp_id, historic=True, model_name=model_name
    )

    if forecast_historic is None:
        logger.debug("Could not find a historic forecast, so will make one")

        forecast_historic = ForecastSQL(
            historic=True,
            forecast_creation_time=datetime.now(timezone.utc),
            location=forecast.location,
            input_data_last_updated=forecast.input_data_last_updated,
            model=forecast.model,
        )
        session.add(forecast_historic)
        session.commit()
    else:
        logger.debug("Found historic forecast")

    return forecast_historic


def upsert(session: Session, model, rows: List[dict]):
    """

    Upsert rows into model

    Insert or Update rows into model.
    This functions checks the primary keys, and if duplicate updates them.

    :param session: sqlalchemy Session
    :param model: the model
    :param rows: the rows we are going to update
    """
    table = model.__table__
    stmt = insert(table)
    primary_keys = [key.name for key in inspect(table).primary_key]
    update_dict = {c.name: c for c in stmt.excluded if not c.primary_key}

    if not update_dict:
        raise ValueError("insert_or_update resulted in an empty update_dict")

    stmt = stmt.on_conflict_do_update(index_elements=primary_keys, set_=update_dict)
    session.execute(stmt, rows)
    session.commit()


def update_forecast_latest(
    forecast: ForecastSQL,
    session: Session,
    forecast_historic: Optional[ForecastSQL] = None,
    model_name: Optional[str] = None,
):
    """
    Update the forecast_values table

    1. Get the Forecast object that is for the latest,
        and don't load children
    2. create ForecastValueLatestSQL objects
    3. upsert them (update and or insert)

    :param forecast:
    :param model_name: the model name to filter on
    :return:
    """

    # 1. get forecast object
    if forecast_historic is None:
        forecast_historic = get_historic_forecast(
            session=session, forecast=forecast, model_name=model_name
        )

    # 2. create forecast value latest
    forecast_values = []
    for forecast_value in forecast.forecast_values:
        forecast_value_latest = change_forecast_value_to_latest(
            forecast_value,
            gsp_id=forecast.location.gsp_id,
            forecast_id=forecast_historic.id,
            model_id=forecast.model_id,
        )
        forecast_values.append(forecast_value_latest.__dict__)

    # upsert forecast values
    upsert(session=session, model=ForecastValueLatestSQL, rows=forecast_values)

    # update forecast creation time
    forecast_historic.forecast_creation_time = datetime.now(tz=timezone.utc)
    session.commit()


def change_forecast_value_to_latest(
    forecast_value: ForecastValueSQL,
    gsp_id: int,
    forecast_id: Optional[int] = None,
    model_id: Optional[int] = None,
) -> ForecastValueLatestSQL:
    """
    Make a ForecastValueLatestSQL from a ForecastValueQL object

    :param forecast_value: forecast value object
    :param gsp_id: gsp id
    :param forecast_id: forecast joining id
    :param model_id: model joining id
    :return: forecast value latest object
    """

    forecast_value_dict = {}
    for v in ["target_time", "expected_power_generation_megawatts", "adjust_mw", "properties"]:
        forecast_value_dict[v] = getattr(forecast_value, v)

    forecast_value_dict["gsp_id"] = gsp_id
    if forecast_id is not None:
        forecast_value_dict["forecast_id"] = forecast_id

    if model_id is not None:
        forecast_value_dict["model_id"] = model_id

    return ForecastValueLatestSQL(**forecast_value_dict)


def update_all_forecast_latest(
    forecasts: List[ForecastSQL],
    session: Session,
    update_national: Optional[bool] = True,
    update_gsp: Optional[bool] = True,
):
    """
    Update all latest forecasts

    :param forecasts: The forecasts that should be updated that should be used to update
        the latest table
    :param session: sqlalmacy session
    :param update_national: Optional (default true), to update the national forecast
    :param update_gsp: Optional (default true), to update all the GSP forecasts
    """

    logger.debug("Getting the earliest forecast target time for the first forecast")
    forecast_values_target_times = [f.target_time for f in forecasts[0].forecast_values]
    model_name = forecasts[0].model.name
    model_version = forecasts[0].model.version
    start_target_time = min(forecast_values_target_times) - timedelta(days=1)
    logger.debug(
        f"First forecast start target time is {min(forecast_values_target_times)} "
        f"so will filter results on {start_target_time} for {model_name}"
    )

    gsp_ids = get_gsp_ids(include_national=update_national, include_gsps=update_gsp)
    if len(gsp_ids) == 0:
        logger.warning(f"Will not be updating any GSPs as {update_national=} and {update_gsp=}")
        return
    logger.debug(f"Will be update GSPs from {min(gsp_ids)} to {max(gsp_ids)}")

    # get all latest forecasts
    logger.debug("Getting all latest forecasts")
    forecasts_historic_all_gsps = get_latest_forecast_for_gsps(
        session=session,
        historic=True,
        preload_children=True,
        gsp_ids=gsp_ids,
        start_target_time=start_target_time,
        model_name=model_name,
    )
    # get all these ids, so we only have to load it once
    historic_gsp_ids = [forecast.location.gsp_id for forecast in forecasts_historic_all_gsps]
    logger.debug(f"Found {len(forecasts_historic_all_gsps)} historic forecasts")

    logger.debug(f"There are {len(forecasts)} forecasts that we will update")

    for forecast in forecasts:
        # chose the correct forecast historic
        logger.debug("Getting gsp")
        gsp_id = forecast.location.gsp_id
        if gsp_id not in gsp_ids:
            logger.debug(
                f"GSP id {gsp_id} will not be updated, "
                f"this is due to the options {update_national=} and {update_gsp=}"
            )
            continue

        logger.debug(f"Updating forecast for gsp_id {gsp_id}")

        forecast_historic = [
            forecast_historic_one_gsps
            for forecast_historic_one_gsps, historic_gsp_id in zip(
                forecasts_historic_all_gsps, historic_gsp_ids
            )
            if historic_gsp_id == gsp_id
        ]
        if len(forecast_historic) == 0:
            forecast_historic = None
            logger.debug(f"Could not find historic, so will be creating one (GSP id{gsp_id})")
        else:
            forecast_historic = forecast_historic[0]

            # update to latest model version
            forecast_historic.model.version = model_version

            logger.debug(f"Found historic for GSP id {gsp_id}")

        update_forecast_latest(
            forecast=forecast,
            session=session,
            forecast_historic=forecast_historic,
            model_name=forecast.model.name,
        )
        session.commit()

    # Delete forecasts older than 3 days from the forecast_latest table
    stmt = delete(ForecastValueLatestSQL).where(
        ForecastValueLatestSQL.target_time < datetime.now(timezone.utc) - timedelta(days=3)
    )
    session.execute(stmt)
    session.commit()


def get_gsp_ids(include_national: bool = True, include_gsps: bool = True) -> List[int]:
    """
    Get list of gsps ids

    :param include_national: option to incldue gsp_id of 0
    :param include_gsps: option to include gsp_ids 1 to N_GSPS
    :return: list of gsps
    """

    gsp_ids = []
    if include_national:
        gsp_ids.append(0)
    if include_gsps:
        gsp_ids = gsp_ids + list(range(1, N_GSP + 1))
    return gsp_ids


def change_forecast_value_to_forecast_last_7_days(
    forecast: ForecastValueSQL,
) -> ForecastValueSevenDaysSQL:
    """
    Make a ForecastValueSevenDaysSQL object from a ForecastValueSQL

    :param forecast: original forecast value
    :return: forecast value seven days sql
    """

    forecast_new = ForecastValueSevenDaysSQL()
    for key in ForecastValueSQL.__table__.columns.keys():
        if hasattr(forecast, key):
            setattr(forecast_new, key, getattr(forecast, key))

    return forecast_new


def add_forecast_last_7_days_and_remove_old_data(
    forecast_values: List[ForecastValueSevenDaysSQL], session: Session
):
    """
    Add forecast values and delete old values

    :param forecast_values:
    :param session:
    :return:
    """

    # add forecast
    session.add_all(forecast_values)

    # remove old data
    now_minus_7_days = datetime.now(tz=timezone.utc) - timedelta(days=7)

    logger.debug(f"Removing data before {now_minus_7_days}")
    query = session.query(ForecastValueSevenDaysSQL)
    query = query.where(ForecastValueSevenDaysSQL.target_time < now_minus_7_days)
    query.delete()



from datetime import datetime, timezone

from nowcasting_datamodel.fake import (
    make_fake_forecast,
    make_fake_forecast_value,
    make_fake_forecasts,
    make_fake_gsp_yields,
    make_fake_input_data_last_updated,
    make_fake_intensity,
    make_fake_location,
    make_fake_national_forecast,
)
from nowcasting_datamodel.models import (
    Forecast,
    ForecastSQL,
    ForecastValue,
    ForecastValueLatestSQL,
    GSPYieldSQL,
    InputDataLastUpdated,
    InputDataLastUpdatedSQL,
    Location,
    LocationSQL,
)
from nowcasting_datamodel.models.forecast import ForecastValueSQL


def test_make_fake_intensity():
    assert make_fake_intensity(datetime(2022, 1, 1)) == 0
    assert make_fake_intensity(datetime(2022, 1, 1, 6)) == 0
    assert 0 < make_fake_intensity(datetime(2022, 1, 1, 9)) < 1
    assert make_fake_intensity(datetime(2022, 1, 1, 12)) == 1
    assert 0 < make_fake_intensity(datetime(2022, 1, 1, 15)) < 1
    assert make_fake_intensity(datetime(2022, 1, 1, 18)) == 0
    assert make_fake_intensity(datetime(2022, 1, 2)) == 0


def test_make_fake_location():
    location_sql: LocationSQL = make_fake_location(1)
    location = Location.from_orm(location_sql)
    _ = Location.to_orm(location)


def test_make_fake_input_data_last_updated():
    input_sql: InputDataLastUpdatedSQL = make_fake_input_data_last_updated()
    input = InputDataLastUpdated.from_orm(input_sql)
    _ = InputDataLastUpdated.to_orm(input)


def test_make_fake_forecast_value():
    target = datetime(2023, 1, 1, tzinfo=timezone.utc)

    forecast_value_sql: ForecastValueSQL = make_fake_forecast_value(target_time=target)
    forecast_value = ForecastValue.from_orm(forecast_value_sql)
    _ = ForecastValue.to_orm(forecast_value)


def test_make_fake_forecast(db_session):
    forecast_sql: ForecastSQL = make_fake_forecast(gsp_id=1, session=db_session)
    forecast = Forecast.from_orm(forecast_sql)
    forecast_sql = Forecast.to_orm(forecast)
    _ = Forecast.from_orm(forecast_sql)

    from sqlalchemy import text

    f = ForecastValueSQL(target_time="2023-01-01 12:00:00", expected_power_generation_megawatts=3)
    db_session.add(f)
    db_session.commit()
    _ = db_session.execute(text("SELECT * FROM forecast_value_2023_01")).all()


def test_make_fake_forecasts(db_session):
    make_fake_forecasts(session=db_session, gsp_ids=[0, 1, 2, 3], add_latest=True)

    assert len(db_session.query(ForecastSQL).all()) > 0
    assert len(db_session.query(ForecastValueLatestSQL).all()) > 0
    assert len(db_session.query(ForecastValueSQL).all()) > 0


def test_make_national_fake_forecast(db_session):
    forecast_sql: ForecastSQL = make_fake_national_forecast(session=db_session)
    forecast = Forecast.from_orm(forecast_sql)
    _ = Forecast.to_orm(forecast)


def test_make_fake_gsp_yields(db_session):
    make_fake_gsp_yields(session=db_session, gsp_ids=list(range(0, 10)))

    assert len(db_session.query(LocationSQL).all()) == 10
    # 2 days for each location, both regimes
    assert len(db_session.query(GSPYieldSQL).all()) == 10 * 48 * 2 * 2



""" Functions used to make fake forecasts"""
from datetime import datetime, time, timedelta, timezone
from typing import List, Optional

import numpy as np
from sqlalchemy.orm import Session

from nowcasting_datamodel.models import (
    InputDataLastUpdatedSQL,
    LocationSQL,
    MetricSQL,
    MetricValueSQL,
    MLModelSQL,
    PVSystemSQL,
    national_gb_label,
)
from nowcasting_datamodel.models.forecast import ForecastSQL, ForecastValueSQL
from nowcasting_datamodel.models.gsp import GSPYieldSQL
from nowcasting_datamodel.read.read import get_location, get_model
from nowcasting_datamodel.read.read_metric import get_datetime_interval
from nowcasting_datamodel.save.update import change_forecast_value_to_latest

# 2 days in the past + 8 hours forward at 30 mins interval
N_FAKE_FORECASTS = (24 * 2 + 8) * 2
TOTAL_MINUTES_IN_ONE_DAY = 24 * 60


def make_fake_location(gsp_id: int) -> LocationSQL:
    """Make fake location with gsp id"""
    return LocationSQL(label=f"GSP_{gsp_id}", gsp_id=gsp_id)


def make_fake_pv_system() -> PVSystemSQL:
    """Make fake location with gsp id"""
    return PVSystemSQL(pv_system_id=1, provider="pvoutput.org", latitude=55, longitude=0)


def make_fake_input_data_last_updated() -> InputDataLastUpdatedSQL:
    """Make fake input data last updated"""
    now = datetime.now(tz=timezone.utc)

    return InputDataLastUpdatedSQL(gsp=now, nwp=now, pv=now, satellite=now)


def make_fake_forecast_value(
    target_time, forecast_maximum: Optional[int] = 1, random_factor: Optional[float] = 1
) -> ForecastValueSQL:
    """Make fake forecast value"""

    intensity = make_fake_intensity(target_time)
    power = forecast_maximum * intensity * random_factor

    return ForecastValueSQL(
        target_time=target_time,
        expected_power_generation_megawatts=power,
        adjust_mw=0.0,
        properties={"10": power * 0.9, "90": power * 1.1},
    )


def make_fake_forecast(
    gsp_id: int,
    session: Session,
    t0_datetime_utc: Optional[datetime] = None,
    forecast_values: Optional[List[ForecastValueSQL]] = None,
    forecast_values_latest: Optional[List[ForecastValueSQL]] = None,
    add_latest: Optional[bool] = False,
    historic: Optional[bool] = False,
    model_name: Optional[str] = "fake_model",
    n_fake_forecasts: Optional[int] = N_FAKE_FORECASTS,
) -> ForecastSQL:
    """Make one fake forecast"""

    if gsp_id == 0:
        # national capacity
        installed_capacity_mw = 13000
    else:
        # gsp capacity (roughly)
        installed_capacity_mw = 10

    location = get_location(
        gsp_id=gsp_id, session=session, installed_capacity_mw=installed_capacity_mw
    )

    model = get_model(name=model_name, session=session, version="0.1.2")
    input_data_last_updated = make_fake_input_data_last_updated()

    if t0_datetime_utc is None:
        t0_datetime_utc = datetime(2024, 1, 1, tzinfo=timezone.utc)

    random_factor = 0.9 + 0.1 * np.random.random()

    # create
    if forecast_values is None:
        forecast_values = []
        # 2 days in the past + 8 hours forward
        for i in range(n_fake_forecasts):
            target_datetime_utc = t0_datetime_utc + timedelta(minutes=i * 30) - timedelta(days=2)
            f = make_fake_forecast_value(
                target_time=target_datetime_utc,
                forecast_maximum=location.installed_capacity_mw,
                random_factor=random_factor,
            )
            forecast_values.append(f)

    if forecast_values_latest is None:
        forecast_values_latest = []

        if add_latest:
            for forecast_value in forecast_values:
                forecast_values_latest.append(
                    change_forecast_value_to_latest(
                        forecast_value, gsp_id=location.gsp_id, model_id=model.id
                    )
                )

    forecast = ForecastSQL(
        model=model,
        forecast_creation_time=t0_datetime_utc,
        location=location,
        input_data_last_updated=input_data_last_updated,
        forecast_values=forecast_values,
        forecast_values_latest=forecast_values_latest,
        historic=historic,
    )

    return forecast


def make_fake_forecasts(
    gsp_ids: List[int],
    session: Session,
    t0_datetime_utc: Optional[datetime] = None,
    forecast_values: Optional[List[ForecastValueSQL]] = None,
    add_latest: Optional[bool] = False,
    historic: Optional[bool] = False,
    model_name: Optional[str] = "fake_model",
    n_fake_forecasts: Optional[int] = N_FAKE_FORECASTS,
) -> List[ForecastSQL]:
    """Make many fake forecast"""
    forecasts = []
    for gsp_id in gsp_ids:
        forecasts.append(
            make_fake_forecast(
                gsp_id=gsp_id,
                t0_datetime_utc=t0_datetime_utc,
                session=session,
                forecast_values=forecast_values,
                add_latest=add_latest,
                historic=historic,
                model_name=model_name,
                n_fake_forecasts=n_fake_forecasts,
            )
        )

    session.add_all(forecasts)

    return forecasts


def make_fake_national_forecast(
    session: Session, t0_datetime_utc: Optional[datetime] = None
) -> ForecastSQL:
    """Make national fake forecast"""
    location = get_location(
        gsp_id=0, session=session, label=national_gb_label, installed_capacity_mw=14000
    )
    model = MLModelSQL(name="fake_model_national", version="0.1.2")
    input_data_last_updated = make_fake_input_data_last_updated()

    if t0_datetime_utc is None:
        t0_datetime_utc = datetime(2024, 1, 1, tzinfo=timezone.utc)

    random_factor = 0.9 + 0.1 * np.random.random()

    # create
    forecast_values = []
    for i in range(N_FAKE_FORECASTS):
        target_datetime_utc = t0_datetime_utc + timedelta(minutes=i * 30) - timedelta(days=2)
        f = make_fake_forecast_value(target_time=target_datetime_utc, random_factor=random_factor)
        forecast_values.append(f)

    forecast = ForecastSQL(
        model=model,
        forecast_creation_time=t0_datetime_utc,
        location=location,
        input_data_last_updated=input_data_last_updated,
        forecast_values=forecast_values,
        historic=False,
    )

    return forecast


def make_fake_gsp_yields_for_one_location(
    gsp_id: int, session: Session, t0_datetime_utc: Optional[datetime] = None
):
    """
    Make GSP yields for one locations

    :param gsp_id: the gsp id you want gsp yields for
    :param session: database sessions
    :param t0_datetime_utc: the time now.
    """

    if gsp_id == 0:
        # national capacity
        installed_capacity_mw = 13000
    else:
        # gsp capacity (roughly)
        installed_capacity_mw = 10

    if t0_datetime_utc is None:
        t0_datetime_utc = datetime(2024, 1, 1, tzinfo=timezone.utc)

    location = get_location(
        session=session, gsp_id=gsp_id, installed_capacity_mw=installed_capacity_mw
    )

    random_factor = 0.9 + 0.1 * np.random.random()

    # make 2 days of fake data
    for i in range(48 * 2):
        datetime_utc = t0_datetime_utc - timedelta(days=2) + timedelta(minutes=i * 30)

        intensity = make_fake_intensity(datetime_utc)
        power = installed_capacity_mw * intensity * random_factor

        gsp_yield = GSPYieldSQL(
            datetime_utc=datetime_utc,
            solar_generation_kw=power,
            regime="in-day",
        )
        gsp_yield.location = location

        session.add(gsp_yield)

        gsp_yield = GSPYieldSQL(
            datetime_utc=datetime_utc, solar_generation_kw=3, regime="day-after"
        )
        gsp_yield.location = location

        session.add(gsp_yield)


def make_fake_intensity(datetime_utc: datetime):
    """
    Make a fake intesnity value based on the time of the day

    :param datetime_utc:
    :return: inteisty, between 0 and 1
    """
    fraction_of_day = (datetime_utc.hour * 60 + datetime_utc.minute) / TOTAL_MINUTES_IN_ONE_DAY
    # use single cos**2 wave for intensity, but set night time to zero
    if (fraction_of_day > 0.25) & (fraction_of_day < 0.75):
        intensity = np.cos(2 * np.pi * fraction_of_day) ** 2
    else:
        intensity = 0.0
    return intensity


def make_fake_gsp_yields(
    session: Session, gsp_ids: List[int], t0_datetime_utc: Optional[datetime] = None
):
    """
    Make GSP yields for many locations

    :param gsp_ids: the gsp ids you want gsp yields for
    :param session: database sessions
    :param t0_datetime_utc: the time now.
    """

    for gsp_id in gsp_ids:
        make_fake_gsp_yields_for_one_location(
            gsp_id=gsp_id, session=session, t0_datetime_utc=t0_datetime_utc
        )


def make_fake_me_latest(session: Session, model_name: str = "fake_model"):
    """Make fake ME latest objects"""
    # create
    metric = MetricSQL(name="Half Hourly ME", description="test")
    datetime_interval = get_datetime_interval(
        session=session,
        start_datetime_utc=datetime(2023, 1, 1),
        end_datetime_utc=datetime(2023, 1, 8),
    )
    location = get_location(
        gsp_id=0, session=session, installed_capacity_mw=14000, label=national_gb_label
    )
    model = MLModelSQL(name=model_name, version="0.1.2")

    session.add_all([location, datetime_interval, metric])

    metric_values = []
    for forecast_horizon in range(0, 60 * 9, 30):
        for minutes in range(0, 60 * 24, 30):
            time_of_day = time(hour=minutes // 60, minute=minutes % 60)
            m = MetricValueSQL(
                value=forecast_horizon * 10000 + minutes,
                time_of_day=time_of_day,
                forecast_horizon_minutes=forecast_horizon,
                number_of_data_points=1,
                datetime_interval=datetime_interval,
                metric=metric,
                location=location,
                model_name=model_name,
                model=model,
            )
            metric_values.append(m)

    return metric_values
